#include "precomp.h"

CInfoTipHandler::CInfoTipHandler(HRESULT*)
    : m_bstrFileName(nullptr), m_lRefCount(1)
{
    ::LockServer();
}

CInfoTipHandler::~CInfoTipHandler()
{
    SysFreeString(m_bstrFileName);
    ::UnlockServer();
}

IFACEMETHODIMP
CInfoTipHandler::QueryInterface(REFIID riid, void** ppvObject)
{
    __try
    {
        *ppvObject = nullptr;

        if(IsEqualIID(IID_IUnknown, riid))
            *ppvObject = GetUnknown();
        else if(IsEqualIID(IID_IQueryInfo, riid))
            *ppvObject = static_cast<IQueryInfo*>(this);
        else if(IsEqualIID(IID_IPersist, riid))
            *ppvObject = static_cast<IPersist*>(this);
        else if(IsEqualIID(IID_IPersistFile, riid))
            *ppvObject = static_cast<IPersistFile*>(this);

        if(*ppvObject)
        {
            static_cast<IUnknown*>(*ppvObject)->AddRef();
            return(S_OK);
        }
    }
    __except((GetExceptionCode() == EXCEPTION_ACCESS_VIOLATION) ? EXCEPTION_EXECUTE_HANDLER : EXCEPTION_CONTINUE_SEARCH)
    {
        return(E_INVALIDARG);
    }

    return(E_NOINTERFACE);
}

IFACEMETHODIMP_(ULONG)
CInfoTipHandler::AddRef()
{
    return(InterlockedIncrement(&m_lRefCount));
}

IFACEMETHODIMP_(ULONG)
CInfoTipHandler::Release()
{
    ULONG ulResult;

    if(0 == (ulResult = InterlockedDecrement(&m_lRefCount)))
        delete this;

    return(ulResult);
}

IFACEMETHODIMP
CInfoTipHandler::GetInfoFlags(DWORD* pdwFlags)
{
    if(!pdwFlags)
        return(E_POINTER);

    *pdwFlags = 0;
    return(S_OK);
}

IFACEMETHODIMP
CInfoTipHandler::GetInfoTip(DWORD, LPWSTR* ppwszTip)
{
    static const PNG png = {{ 0x89, 0x50, 0x4E, 0x47, 0x0D, 0x0A, 0x1A, 0x0A }, { 0x0D000000, 0x52444849 }};
    BITMAPINFOHEADER    bmih;
    OVERLAPPED          ov;
    ICONDIR             icondir;
    HRESULT             hr = S_OK;
    LPWSTR              pwszTip = nullptr;
    HANDLE              hFile;
    PNG*                pPng;
    WORD                wIndex;

    if(!ppwszTip)
        return(E_POINTER);

    *ppwszTip = nullptr;

    if(!m_bstrFileName)
        return(OLE_E_BLANK);

    if(INVALID_HANDLE_VALUE == (hFile = CreateFileW(m_bstrFileName, GENERIC_READ, FILE_SHARE_READ, nullptr, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS, nullptr)))
        return(HRESULT_FROM_WIN32(GetLastError()));

    __try
    {
        ZeroMemory(&ov, sizeof(ov));

        if(FAILED(hr = ReadFromFile(hFile, &icondir, sizeof(icondir), &ov)))
            __leave;

        if((icondir.idReserved != 0) || ((icondir.idType != IMAGE_ICON) && (icondir.idType != IMAGE_CURSOR)) || (icondir.idCount == 0))
        {
            hr = __HRESULT_FROM_WIN32(ERROR_FILE_CORRUPT);
            __leave;
        }

        if(FAILED(hr = GetTipHeader(&pwszTip, icondir.idCount)))
            __leave;

        for(wIndex = 0;;)
        {
            ov.Offset = icondir.idEntries[0].dwImageOffset;

            if(FAILED(hr = ReadFromFile(hFile, &bmih, sizeof(bmih), &ov)))
                __leave;

            if(!memcmp(&png, &bmih, offsetof(PNG, hdr) + sizeof(CHUNKHEADER)))
            {
                pPng = reinterpret_cast<PNG*>(&bmih);

                if(FAILED(hr = AppendImageInfo(&pwszTip, wIndex + 1, my_ntohl(pPng->ihdr.Width), my_ntohl(pPng->ihdr.Height), pPng->ihdr.BitDepth, (ULONGLONG)(1ULL << (ULONGLONG)pPng->ihdr.BitDepth))))
                    __leave;
            }
            else
            {
                if(FAILED(hr = AppendImageInfo(&pwszTip, wIndex + 1, icondir.idEntries[0].bWidth, icondir.idEntries[0].bHeight, bmih.biBitCount, (ULONGLONG)(1ULL << (ULONGLONG)bmih.biBitCount))))
                    __leave;
            }

            if(++wIndex >= icondir.idCount)
                break;

            #pragma warning(suppress: 6385)
            ov.Offset = static_cast<DWORD>(offsetof(ICONDIR, idEntries[wIndex]));

            if(FAILED(hr = ReadFromFile(hFile, &icondir.idEntries[0], sizeof(icondir.idEntries[0]), &ov)))
                __leave;
        }

        *ppwszTip = pwszTip;
        pwszTip = nullptr;
    }
    __finally
    {
        CoTaskMemFree(pwszTip);
        CloseHandle(hFile);
    }

    return(hr);
}

IFACEMETHODIMP
CInfoTipHandler::GetClassID(CLSID* pClassID)
{
    if(!pClassID)
        return(E_POINTER);

    *pClassID = CLSID_InfoTipHandler;
    return(S_OK);
}

IFACEMETHODIMP
CInfoTipHandler::IsDirty()
{
    return(S_FALSE);
}

IFACEMETHODIMP
CInfoTipHandler::Load(LPCOLESTR pszFileName, DWORD)
{
    BSTR bstr;

    if(!pszFileName)
        return(E_POINTER);

    if(nullptr == (bstr = SysAllocString(pszFileName)))
        return(E_OUTOFMEMORY);

    SysFreeString(m_bstrFileName);
    m_bstrFileName = bstr;
    return(S_OK);
}

IFACEMETHODIMP
CInfoTipHandler::Save(LPCOLESTR, BOOL)
{
    return(E_NOTIMPL);
}

IFACEMETHODIMP
CInfoTipHandler::SaveCompleted(LPCOLESTR)
{
    return(S_OK);
}

IFACEMETHODIMP
CInfoTipHandler::GetCurFile(LPOLESTR* ppszFileName)
{
    IMalloc* pMalloc = nullptr;
    HRESULT  hr;

    if(!m_bstrFileName)
        return(E_FAIL);

    if(!ppszFileName)
        return(E_POINTER);

    if(FAILED(hr = CoGetMalloc(1, &pMalloc)))
        return(hr);

    *ppszFileName = static_cast<LPOLESTR>(pMalloc->Alloc(SysStringByteLen(m_bstrFileName) + sizeof(OLECHAR)));
    pMalloc->Release();

    if(!*ppszFileName)
        return(E_OUTOFMEMORY);

    wcscpy(*ppszFileName, m_bstrFileName);
    return(S_OK);
}

HRESULT
CInfoTipHandler::GetTipHeader(LPWSTR* ppwszHdr, UINT64 dwNumImages)
{
    CThreadLocale   loc(MAKELCID(MAKELANGID(PRIMARYLANGID(CThreadLocale::getLANGID()), SUBLANG_DEFAULT), SORT_DEFAULT));
    HRESULT         hr;
    LPWSTR          pwsz = nullptr;

    if(!FormatMessageW(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_HMODULE | FORMAT_MESSAGE_ARGUMENT_ARRAY, g_hInst, (dwNumImages == 1) ? MSGID_HEADER_SINGLE : MSGID_HEADER_MULTIPLE, 0, reinterpret_cast<LPWSTR>(&pwsz), 0, reinterpret_cast<va_list*>(&dwNumImages)))
        return(HRESULT_FROM_WIN32(GetLastError()));

    if(!pwsz)
        return(E_UNEXPECTED);

    const size_t len = wcslen(pwsz) + 1;

    if(nullptr == (*ppwszHdr = static_cast<LPWSTR>(CoTaskMemAlloc(len * sizeof(WCHAR)))))
    {
        hr = E_OUTOFMEMORY;
    }
    else
    {
        wcscpy(*ppwszHdr, pwsz);
        hr = S_OK;
    }

    LocalFree(pwsz);
    return(hr);
}

HRESULT
CInfoTipHandler::AppendImageInfo(__inout LPWSTR* ppwszTip, ...)
{
    CThreadLocale   loc(MAKELCID(MAKELANGID(PRIMARYLANGID(CThreadLocale::getLANGID()), SUBLANG_DEFAULT), SORT_DEFAULT));
    HRESULT         hr = S_OK;
    va_list         lst;
    LPWSTR          pwsz = nullptr, ptmp;

    va_start(lst, ppwszTip);

    if(!FormatMessageW(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_HMODULE, g_hInst, MSGID_IMAGEINFO, 0, reinterpret_cast<LPWSTR>(&pwsz), 0, &lst))
        hr = HRESULT_FROM_WIN32(GetLastError());

    va_end(lst);

    if(FAILED(hr))
        return(hr);

    if(!pwsz)
        return(E_UNEXPECTED);

    const size_t len = wcslen(*ppwszTip) + wcslen(pwsz) + 1;

    if(nullptr == (ptmp = static_cast<LPWSTR>(CoTaskMemRealloc(*ppwszTip, len * sizeof(WCHAR)))))
    {
        hr = E_OUTOFMEMORY;
    }
    else
    {
        *ppwszTip = ptmp;
        wcscat(*ppwszTip, pwsz);
    }

    LocalFree(pwsz);
    return(hr);
}
